"""Utils for SLURM backend."""

import pprint
import re
from typing import Dict

from waldur_site_agent.backends import logger
from waldur_site_agent.backends.exceptions import BackendError

from .backend import SlurmBackend

UNIT_PATTERN = re.compile(r"(\d+)([KMGTP]?)")

UNITS: Dict[str, int] = {
    "K": 2**10,
    "M": 2**20,
    "G": 2**30,
    "T": 2**40,
}


def parse_int(value: str) -> int:
    """Converts human-readable integers to machine-readable ones.

    Example: 5K to 5000.
    """
    match = re.match(UNIT_PATTERN, value)
    if not match:
        return 0
    value_ = int(match.group(1))
    unit = match.group(2)
    factor = UNITS[unit] if unit else 1
    return factor * value_


def diagnostics(slurm_backend: SlurmBackend) -> bool:
    """Runs diagnostics for SLURM cluster."""
    default_account_name = slurm_backend.backend_settings["default_account"]

    format_string = "{:<30} = {:<10}"
    logger.info(
        format_string.format(
            "SLURM customer prefix", slurm_backend.backend_settings["customer_prefix"]
        )
    )
    logger.info(
        format_string.format(
            "SLURM project prefix", slurm_backend.backend_settings["project_prefix"]
        )
    )
    logger.info(
        format_string.format(
            "SLURM allocation prefix", slurm_backend.backend_settings["allocation_prefix"]
        )
    )
    logger.info(format_string.format("SLURM default account", default_account_name))
    logger.info("")

    logger.info("SLURM tres components:\n%s\n", pprint.pformat(slurm_backend.backend_components))

    try:
        slurm_version_info = slurm_backend.client._execute_command(
            ["-V"], "sinfo", immediate=False, parsable=False
        )
        logger.info("Slurm version: %s", slurm_version_info.strip())
    except BackendError as err:
        logger.error("Unable to fetch SLURM info, reason: %s", err)
        return False

    try:
        slurm_backend.ping(raise_exception=True)
        logger.info("SLURM cluster ping is successful")
    except BackendError as err:
        logger.error("Unable to ping SLURM cluster, reason: %s", err)

    tres = slurm_backend.list_components()
    logger.info("Available tres in the cluster: %s", ",".join(tres))

    default_account = slurm_backend.client.get_account(default_account_name)
    if default_account is None:
        logger.error("There is no account %s in the cluster", default_account)
        return False
    logger.info('Default parent account "%s" is in place', default_account_name)
    logger.info("")

    return True


def convert_slurm_units_to_waldur_ones(slurm_tres: Dict, units: Dict, to_int: bool = False) -> Dict:
    """Converts SLURM computing units to Waldur ones."""
    converted_units = {}

    for tres_name, value in units.items():
        converted_value = value
        converted_value = converted_value / slurm_tres[tres_name].get("unit_factor", 1)

        if to_int:
            converted_units[tres_name] = int(converted_value)
        else:
            converted_units[tres_name] = round(converted_value, 2)

    return converted_units



"""Agent responsible for membership control."""

import traceback
from time import sleep
from typing import Dict, List, Set

from waldur_site_agent import common_utils
from waldur_site_agent.backends import logger
from waldur_site_agent.backends.structures import Resource
from waldur_site_agent.processors import OfferingBaseProcessor

from . import (
    MARKETPLACE_SLURM_OFFERING_TYPE,
    WALDUR_SITE_AGENT_MEMBERSHIP_SYNC_PERIOD_MINUTES,
    Offering,
    WaldurAgentConfiguration,
)


class OfferingMembershipProcessor(OfferingBaseProcessor):
    """Class for an offering processing.

    Processes related resources and reports membership data to Waldur.
    """

    def process_offering(self) -> None:
        """Processes offering and reports resources usage to Waldur."""
        logger.info(
            "Processing offering %s (%s)",
            self.offering.name,
            self.offering.uuid,
        )
        self._print_current_user()

        waldur_offering = self.waldur_rest_client._get_offering(self.offering.uuid)
        common_utils.extend_backend_components(self.offering, waldur_offering["components"])

        waldur_resources = self.waldur_rest_client.filter_marketplace_provider_resources(
            {
                "offering_uuid": self.offering.uuid,
                "state": "OK",
                "field": [
                    "backend_id",
                    "uuid",
                    "name",
                    "resource_uuid",
                    "offering_type",
                    "restrict_member_access",
                    "downscaled",
                    "paused",
                ],
            }
        )

        if len(waldur_resources) == 0:
            logger.info("No resources to process")
            return

        offering_type = waldur_resources[0].get("offering_type", "")

        waldur_resources_info = [
            Resource(
                name=resource_data["name"],
                backend_id=resource_data["backend_id"],
                marketplace_uuid=resource_data["uuid"],
                backend_type=self.offering.backend_type,
                marketplace_scope_uuid=resource_data["resource_uuid"],
                restrict_member_access=resource_data.get("restrict_member_access", False),
                downscaled=resource_data.get("downscaled", False),
                paused=resource_data.get("paused", False),
            )
            for resource_data in waldur_resources
        ]

        resource_report = self.resource_backend.pull_resources(waldur_resources_info)

        self._process_resources(resource_report, offering_type)

    def _sync_slurm_resource_users(
        self,
        backend_resource: Resource,
    ) -> None:
        """Syncs users for the resource between SLURM cluster and Waldur."""
        # This method is currently implemented for SLURM backend only
        logger.info("Syncing user list for resource %s", backend_resource.name)
        usernames = backend_resource.users
        local_usernames = set(usernames)
        logger.info("The usernames from the backend: %s", ", ".join(local_usernames))

        # Offering users sync
        # The service fetches offering users from Waldur and pushes them to the cluster
        # If an offering user is not in the team anymore, it will be removed from the backend
        logger.info("Synching offering users")
        team = self.waldur_rest_client.marketplace_provider_resource_get_team(
            backend_resource.marketplace_uuid
        )
        team_user_uuids = {user["uuid"] for user in team}

        offering_users = self.waldur_rest_client.list_remote_offering_users(
            {
                "offering_uuid": self.offering.uuid,
                "is_restricted": False,
            }
        )

        if backend_resource.restrict_member_access:
            # The idea is to remove the existing associations in both sides
            # and avoid creation of new associations
            logger.info("Resource restricted for members, removing all the existing associations")
            existing_offering_user_usernames: Set[str] = {
                offering_user["username"]
                for offering_user in offering_users
                if offering_user["username"] in local_usernames
                and offering_user["user_uuid"] in team_user_uuids
            }

            common_utils.delete_associations_from_waldur_allocation(
                self.waldur_rest_client, backend_resource, existing_offering_user_usernames
            )
            self.resource_backend.remove_users_from_account(
                backend_resource.backend_id, existing_offering_user_usernames
            )
            return

        new_offering_user_usernames: Set[str] = {
            offering_user["username"]
            for offering_user in offering_users
            if offering_user["username"] not in local_usernames
            and offering_user["user_uuid"] in team_user_uuids
        }

        stale_offering_user_usernames: Set[str] = {
            offering_user["username"]
            for offering_user in offering_users
            if offering_user["username"] in local_usernames
            and offering_user["user_uuid"] not in team_user_uuids
        }

        common_utils.create_associations_for_waldur_allocation(
            self.waldur_rest_client, backend_resource, new_offering_user_usernames
        )

        self.resource_backend.add_users_to_resource(
            backend_resource.backend_id,
            new_offering_user_usernames,
            homedir_umask=self.offering.backend_settings.get("homedir_umask", "0700"),
        )

        common_utils.delete_associations_from_waldur_allocation(
            self.waldur_rest_client, backend_resource, stale_offering_user_usernames
        )

        self.resource_backend.remove_users_from_account(
            backend_resource.backend_id,
            stale_offering_user_usernames,
        )

    def _process_resources(
        self,
        resource_report: Dict[str, Resource],
        offering_type: str,
    ) -> None:
        """Sync membership data for the resource."""
        # Push data to Mastermind using REST client

        for resource_backend_id, backend_resource in resource_report.items():
            logger.info("-" * 30)
            try:
                logger.info("Processing %s", resource_backend_id)
                # Sync users
                if offering_type == MARKETPLACE_SLURM_OFFERING_TYPE:
                    self._sync_slurm_resource_users(backend_resource)

                if backend_resource.paused:
                    logger.info("The resource pausing is requested, processing it")
                    pausing_done = self.resource_backend.pause_resource(backend_resource.backend_id)
                    if pausing_done:
                        logger.info("The pausing is successfully completed")
                    else:
                        logger.warning("The pausing is not done")
                elif backend_resource.downscaled:
                    logger.info("The resource downscaling is requested, processing it")
                    downscaling_done = self.resource_backend.downscale_resource(
                        backend_resource.backend_id
                    )
                    if downscaling_done:
                        logger.info("The downscaling is successfully completed")
                    else:
                        logger.warning("The downscaling is not done")
                else:
                    logger.info(
                        "The resource is not downscaled or paused, "
                        "resetting the QoS to the default one"
                    )
                    restoring_done = self.resource_backend.restore_resource(
                        backend_resource.backend_id
                    )
                    if restoring_done:
                        logger.info("The restoring is successfully completed")
                    else:
                        logger.info("The restoring skipped")

                resource_metadata = self.resource_backend.get_resource_metadata(
                    backend_resource.backend_id
                )
                self.waldur_rest_client.marketplace_provider_resource_set_backend_metadata(
                    backend_resource.marketplace_uuid, resource_metadata
                )
            except Exception as e:
                logger.exception(
                    "Waldur REST client error while processing allocation %s: %s",
                    resource_backend_id,
                    e,
                )
                error_traceback = traceback.format_exc()
                common_utils.mark_waldur_resources_as_erred(
                    self.waldur_rest_client,
                    [backend_resource],
                    error_details={
                        "error_message": str(e),
                        "error_traceback": error_traceback,
                    },
                )


def process_offerings(waldur_offerings: List[Offering], user_agent: str = "") -> None:
    """Processes list of offerings."""
    logger.info("Number of offerings to process: %s", len(waldur_offerings))
    for offering in waldur_offerings:
        try:
            processor = OfferingMembershipProcessor(offering, user_agent)
            processor.process_offering()
        except Exception as e:
            logger.exception("The application crashed due to the error: %s", e)


def start(configuration: WaldurAgentConfiguration) -> None:
    """Starts the main loop for offering processing."""
    logger.info("Synching data to Waldur")
    while True:
        try:
            process_offerings(configuration.waldur_offerings, configuration.waldur_user_agent)
        except Exception as e:
            logger.exception("The application crashed due to the error: %s", e)
        sleep(WALDUR_SITE_AGENT_MEMBERSHIP_SYNC_PERIOD_MINUTES * 60)



"""Common structures shared between different backends."""

from dataclasses import dataclass, field
from typing import Dict, List


@dataclass
class Account:
    """Account model for SLURM and MOAB."""

    name: str = ""
    description: str = ""
    organization: str = ""


@dataclass
class Association:
    """Association between a user and an account in SLURM or MOAB."""

    account: str = ""
    user: str = ""
    value: int = 0


@dataclass
class Resource:
    """Common resource structure."""

    backend_type: str = ""
    name: str = ""
    marketplace_uuid: str = ""
    marketplace_scope_uuid: str = ""
    backend_id: str = ""
    limits: Dict = field(default_factory=dict)
    restrict_member_access: bool = False
    downscaled: bool = False
    paused: bool = False
    users: List[str] = field(default_factory=list)
    usage: Dict = field(default_factory=dict)
    metadata: Dict = field(default_factory=dict)
    state: str = ""



"""Generic client class."""

import abc
import subprocess
from typing import Dict, List, Optional

from waldur_site_agent.backends import logger
from waldur_site_agent.backends.exceptions import (
    BackendError,
)
from waldur_site_agent.backends.structures import Account, Association


class BaseClient:
    """Generic cli-client for a backend communication."""

    def execute_command(self, command: List[str]) -> str:
        """Execute command on backend."""
        try:
            logger.debug("Executing command: %s", " ".join(command))
            return subprocess.check_output(command, stderr=subprocess.STDOUT, encoding="utf-8")
        except subprocess.CalledProcessError as e:
            logger.exception('Failed to execute command "%s".', command)
            stdout = e.output or ""
            lines = stdout.splitlines()
            if len(lines) > 0 and lines[0].startswith("Warning: Permanently added"):
                lines = lines[1:]
            stdout = "\n".join(lines)
            raise BackendError(stdout) from e

    @abc.abstractmethod
    def list_accounts(self) -> List[Account]:
        """Get accounts list."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_account(self, name: str) -> Optional[Account]:
        """Get account info."""
        raise NotImplementedError

    @abc.abstractmethod
    def create_account(
        self, name: str, description: str, organization: str, parent_name: Optional[str] = None
    ) -> str:
        """Create account in the cluster."""
        raise NotImplementedError

    @abc.abstractmethod
    def delete_account(self, name: str) -> str:
        """Delete account from the cluster."""
        raise NotImplementedError

    @abc.abstractmethod
    def set_resource_limits(self, account: str, limits_dict: Dict[str, int]) -> Optional[str]:
        """Set account limits."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_association(self, user: str, account: str) -> Optional[Association]:
        """Get association between user and account."""
        raise NotImplementedError

    @abc.abstractmethod
    def create_association(
        self, username: str, account: str, default_account: Optional[str] = None
    ) -> str:
        """Create association between user and account."""
        raise NotImplementedError

    @abc.abstractmethod
    def delete_association(self, username: str, account: str) -> str:
        """Delete association between user and account."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_usage_report(self, accounts: List[str]) -> List:
        """Get usage records."""
        raise NotImplementedError

    @abc.abstractmethod
    def list_account_users(self, account: str) -> List[str]:
        """Get account users."""
        raise NotImplementedError


class UnknownClient(BaseClient):
    """Unknown cli-client for a backend communication."""

    def list_accounts(self) -> List[Account]:
        """Get accounts list."""
        return []

    def get_account(self, _: str) -> Optional[Account]:
        """Get account info."""
        return None

    def create_account(
        self, name: str, description: str, organization: str, parent_name: Optional[str] = None
    ) -> str:
        """Create account in the cluster."""
        del description, organization, parent_name
        return name

    def delete_account(self, name: str) -> str:
        """Delete account from the cluster."""
        return name

    def set_resource_limits(self, account: str, limits_dict: Dict[str, int]) -> Optional[str]:
        """Set account limits."""
        del account, limits_dict
        return ""

    def get_association(self, user: str, account: str) -> Optional[Association]:
        """Get association between user and account."""
        del user, account
        return None

    def create_association(
        self, username: str, account: str, default_account: Optional[str] = None
    ) -> str:
        """Create association between user and account."""
        del account, default_account
        return username

    def delete_association(self, username: str, account: str) -> str:
        """Delete association between user and account."""
        del account
        return username

    def get_usage_report(self, accounts: List[str]) -> List:
        """Get usages records."""
        return accounts

    def list_account_users(self, _: str) -> List[str]:
        """Get account users."""
        return []



"""Agent responsible for usage and limits reporting."""

import datetime
import traceback
from time import sleep
from typing import Dict, List

from waldur_client import (
    ComponentUsage,
)

from waldur_site_agent.backends import logger, utils
from waldur_site_agent.backends.structures import Resource
from waldur_site_agent.processors import OfferingBaseProcessor

from . import (
    MARKETPLACE_SLURM_OFFERING_TYPE,
    WALDUR_SITE_AGENT_REPORT_PERIOD_MINUTES,
    Offering,
    WaldurAgentConfiguration,
    common_utils,
)


class OfferingReportProcessor(OfferingBaseProcessor):
    """Class for an offering processing.

    Processes related resource and reports computing data to Waldur.
    """

    def process_offering(self) -> None:
        """Processes offering and reports resources usage to Waldur."""
        logger.info(
            "Processing offering %s (%s)",
            self.offering.name,
            self.offering.uuid,
        )
        self._print_current_user()

        waldur_offering = self.waldur_rest_client._get_offering(self.offering.uuid)
        common_utils.extend_backend_components(self.offering, waldur_offering["components"])

        waldur_resources = self.waldur_rest_client.filter_marketplace_provider_resources(
            {
                "offering_uuid": self.offering.uuid,
                "state": ["OK", common_utils.RESOURCE_ERRED_STATE],
                "field": ["backend_id", "uuid", "name", "offering_type", "state"],
            }
        )

        if len(waldur_resources) == 0:
            logger.info("No resources to process")
            return

        offering_type = waldur_resources[0].get("offering_type", "")

        waldur_resources_info = [
            Resource(
                name=resource_data["name"],
                backend_id=resource_data["backend_id"],
                marketplace_uuid=resource_data["uuid"],
                backend_type=self.offering.backend_type,
                state=resource_data["state"],
            )
            for resource_data in waldur_resources
        ]

        resource_report = self.resource_backend.pull_resources(waldur_resources_info)

        # TODO: make generic
        if offering_type == MARKETPLACE_SLURM_OFFERING_TYPE:
            # Allocations existing in Waldur but missing in SLURM cluster
            missing_resources = [
                Resource(
                    marketplace_uuid=resource_info["uuid"],
                    backend_id=resource_info["backend_id"],
                )
                for resource_info in waldur_resources
                if resource_info["backend_id"] not in set(resource_report.keys())
                and resource_info["state"] != common_utils.RESOURCE_ERRED_STATE
            ]
            logger.info("Number of missing resources %s", len(missing_resources))
            if len(missing_resources) > 0:
                common_utils.mark_waldur_resources_as_erred(
                    self.waldur_rest_client,
                    missing_resources,
                    {"error_message": "The resource is missing on the backend"},
                )

        self._process_resources(resource_report)

    def _submit_total_usage_for_resource(
        self,
        backend_resource: Resource,
        total_usage: Dict[str, float],
        waldur_components: List[Dict],
    ) -> None:
        """Reports total usage for a backend resource to Waldur."""
        logger.info("Setting usages: %s", total_usage)
        resource_uuid = backend_resource.marketplace_uuid
        plan_periods = self.waldur_rest_client.marketplace_provider_resource_get_plan_periods(
            resource_uuid
        )

        if len(plan_periods) == 0:
            logger.warning(
                "A corresponding ResourcePlanPeriod for resource %s was not found",
                backend_resource.name,
            )
            return

        plan_period = plan_periods[0]
        component_types = [component["type"] for component in waldur_components]
        missing_components = set(total_usage) - set(component_types)

        if missing_components:
            logger.warning(
                "The following components are not found in Waldur: %s",
                ", ".join(missing_components),
            )

        usage_objects = [
            ComponentUsage(type=component, amount=amount)
            for component, amount in total_usage.items()
            if component in component_types
        ]
        self.waldur_rest_client.create_component_usages(plan_period["uuid"], usage_objects)

    def _submit_user_usage_for_resource(
        self,
        username: str,
        user_usage: Dict[str, float],
        waldur_component_usages: List[Dict],
    ) -> None:
        """Reports per-user usage for a backend resource to Waldur."""
        logger.info("Setting usages for %s", username)
        component_usage_types = [
            component_usage["type"] for component_usage in waldur_component_usages
        ]
        missing_components = set(user_usage) - set(component_usage_types)

        if missing_components:
            logger.warning(
                "The following components are not found in Waldur: %s",
                ", ".join(missing_components),
            )

        offering_users = self.waldur_rest_client.list_remote_offering_users(
            {"username": username, "query": self.offering.uuid}
        )
        offering_user_uuid = None

        if len(offering_users) > 0:
            offering_user_uuid = offering_users[0]["uuid"]

        for component_usage in waldur_component_usages:
            component_type = component_usage["type"]
            usage = user_usage[component_type]
            logger.info(
                "Submitting usage for username %s: %s -> %s",
                username,
                component_type,
                usage,
            )
            self.waldur_rest_client.create_component_user_usage(
                component_usage["uuid"], usage, username, offering_user_uuid
            )

    def _process_resources(
        self,
        resource_report: Dict[str, Resource],
    ) -> None:
        """Processes usage report for the resource."""
        waldur_offering = self.waldur_rest_client._get_offering(self.offering.uuid)
        month_start = utils.month_start(datetime.datetime.now()).date()

        # TODO: this part is not generic yet, rather SLURM-specific
        for resource_backend_id, backend_resource in resource_report.items():
            try:
                logger.info("Processing %s", resource_backend_id)
                usages: Dict[str, Dict[str, float]] = backend_resource.usage

                # Set resource state OK if it is erred
                if backend_resource.state == common_utils.RESOURCE_ERRED_STATE:
                    self.waldur_rest_client.marketplace_provider_resource_set_as_ok(
                        backend_resource.marketplace_uuid
                    )

                # Submit usage
                total_usage = usages.pop("TOTAL_ACCOUNT_USAGE")
                self._submit_total_usage_for_resource(
                    backend_resource,
                    total_usage,
                    waldur_offering["components"],
                )

                # Skip the following actions if the dict is empty
                if not usages:
                    continue

                waldur_component_usages = self.waldur_rest_client.list_component_usages(
                    backend_resource.marketplace_uuid, date_after=month_start
                )

                logger.info("Setting per-user usages")
                for username, user_usage in usages.items():
                    self._submit_user_usage_for_resource(
                        username, user_usage, waldur_component_usages
                    )
            except Exception as e:
                logger.exception(
                    "Waldur REST client error while processing allocation %s: %s",
                    resource_backend_id,
                    e,
                )
                error_traceback = traceback.format_exc()
                common_utils.mark_waldur_resources_as_erred(
                    self.waldur_rest_client,
                    [backend_resource],
                    error_details={
                        "error_message": str(e),
                        "error_traceback": error_traceback,
                    },
                )


def process_offerings(waldur_offerings: List[Offering], user_agent: str = "") -> None:
    """Processes list of offerings."""
    logger.info("Number of offerings to process: %s", len(waldur_offerings))
    for offering in waldur_offerings:
        try:
            processor = OfferingReportProcessor(offering, user_agent)
            processor.process_offering()
        except Exception as e:
            logger.exception("The application crashed due to the error: %s", e)


def start(configuration: WaldurAgentConfiguration) -> None:
    """Starts the main loop for offering processing."""
    logger.info("Synching data to Waldur")
    while True:
        try:
            process_offerings(configuration.waldur_offerings, configuration.waldur_user_agent)
        except Exception as e:
            logger.exception("The application crashed due to the error: %s", e)
        sleep(WALDUR_SITE_AGENT_REPORT_PERIOD_MINUTES * 60)



"""Module for abstract offering processor."""

import abc

from waldur_client import WaldurClient

from waldur_site_agent import Offering, common_utils
from waldur_site_agent.backends import BackendType
from waldur_site_agent.backends.exceptions import BackendError


class OfferingBaseProcessor(abc.ABC):
    """Abstract class for an offering processing."""

    def __init__(self, offering: Offering, user_agent: str = "") -> None:
        """Constructor."""
        self.offering: Offering = offering
        self.waldur_rest_client: WaldurClient = WaldurClient(
            offering.api_url, offering.api_token, user_agent
        )
        self.resource_backend = common_utils.get_backend_for_offering(offering)
        if self.resource_backend.backend_type == BackendType.UNKNOWN.value:
            raise BackendError(f"Unable to create backend for {self.offering}")

    def _print_current_user(self) -> None:
        current_user = self.waldur_rest_client.get_current_user()
        common_utils.print_current_user(current_user)

    @abc.abstractmethod
    def process_offering(self) -> None:
        """Pulls data form Mastermind using REST client and creates objects on the backend."""



"""Moab-specific backend classes and functions."""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Tuple

from waldur_site_agent.backends import BackendType, logger
from waldur_site_agent.backends import utils as backend_utils
from waldur_site_agent.backends.backend import BaseBackend
from waldur_site_agent.backends.exceptions import BackendError
from waldur_site_agent.backends.moab_backend.client import MoabClient

if TYPE_CHECKING:
    from waldur_site_agent.backends.moab_backend.parser import MoabReportLine


class MoabBackend(BaseBackend):
    """MOAB backend class."""

    def __init__(self, moab_settings: Dict, moab_components: Dict[str, Dict]) -> None:
        """Init backend data and creates a corresponding client."""
        super().__init__(moab_settings, moab_components)
        self.backend_type = BackendType.MOAB.value
        self.client = MoabClient()
        self.backend_components["deposit"]["unit_factor"] = 1

    def ping(self, raise_exception: bool = False) -> bool:
        """Check if MOAB is online."""
        try:
            self.client.list_accounts()
        except BackendError as err:
            if raise_exception:
                raise
            logger.info("Error: %s", err)
            return False
        else:
            return True

    def list_components(self) -> List[str]:
        """Return deposit component."""
        return ["deposit"]

    def _get_usage_report(self, accounts: List[str]) -> Dict:
        """Get usage report."""
        report: Dict[str, Dict[str, Dict[str, int]]] = {}
        lines: List[MoabReportLine] = self.client.get_usage_report(accounts)

        for line in lines:
            report.setdefault(line.account, {}).setdefault(line.user, {})
            user_usage_existing = report[line.account][line.user]
            user_usage_new = backend_utils.sum_dicts([user_usage_existing, line.usages])
            report[line.account][line.user] = user_usage_new

        for account_usage in report.values():
            usages_per_user = list(account_usage.values())
            total = backend_utils.sum_dicts(usages_per_user)
            account_usage["TOTAL_ACCOUNT_USAGE"] = total

        return report

    def downscale_resource(self, account: str) -> bool:
        """Temporary placeholder."""
        del account
        return False

    def pause_resource(self, account: str) -> bool:
        """Temporary placeholder."""
        del account
        return False

    def restore_resource(self, account: str) -> bool:
        """Temporary placeholder."""
        del account
        return False

    def get_resource_metadata(self, _: str) -> dict:
        """Temporary placeholder."""
        return {}

    def _collect_limits(
        self, waldur_resource: Dict[str, Dict]
    ) -> Tuple[Dict[str, int], Dict[str, int]]:
        """Collect deposit limit only with no conversion."""
        deposit_limit = {"deposit": waldur_resource["limits"]["deposit"]}
        return deposit_limit, deposit_limit



"""CLI-client for MOAB."""

from __future__ import annotations

from typing import Dict, List, Optional

from waldur_site_agent.backends import base, exceptions, logger
from waldur_site_agent.backends import utils as backend_utils
from waldur_site_agent.backends.structures import Account, Association

from .parser import MoabReportLine


class MoabClient(base.BaseClient):
    """This class implements Python client for MOAB.

    See also MOAB Accounting Manager 9.1.1 Administrator Guide
    http://docs.adaptivecomputing.com/9-1-1/MAM/help.htm
    """

    def list_accounts(self) -> List[Account]:
        """Return list of accounts in MOAB."""
        output = self.execute_command(
            "mam-list-accounts --raw --quiet --show Name,Description,Organization".split()
        )
        return [self._parse_account(line) for line in output.splitlines() if "|" in line]

    def _parse_account(self, line: str) -> Account:
        parts = line.split("|")
        return Account(name=parts[0], description=parts[1], organization=parts[2])

    def _get_fund_id(self, account: str) -> Optional[int]:
        command_fund = f"mam-list-funds --raw --quiet -a {account} --show Id"
        fund_output = self.execute_command(command_fund.split())
        fund_data = fund_output.splitlines()

        if len(fund_data) == 0:
            logger.warning("No funds were found for account %s", account)
            return None

        # Assuming an account has only one fund
        fund_id_str = fund_data[0].strip()

        return int(fund_id_str)

    def get_account(self, name: str) -> Account | None:
        """Get MOAB account info."""
        command = (
            "mam-list-accounts --raw --quiet --show Name,Description,Organization -a %s" % name
        )
        output = self.execute_command(command.split())
        lines = [line for line in output.splitlines() if "|" in line]
        if len(lines) == 0:
            return None
        return self._parse_account(lines[0])

    def create_account(
        self, name: str, description: str, organization: str, parent_name: Optional[str] = None
    ) -> str:
        """Create account in MOAB."""
        del parent_name
        command_account = f'mam-create-account -a {name} -d "{description}" -o {organization}'
        self.execute_command(command_account.split())

        logger.info("Creating fund for the account")
        command_fund = f"mam-create-fund -a {name}"
        return self.execute_command(command_fund.split())

    def delete_account(self, name: str) -> str:
        """Delete account from MOAB."""
        command_account = "mam-delete-account -a %s" % name
        self.execute_command(command_account.split())

        fund_id = self._get_fund_id(name)

        if fund_id is None:
            logger.warning("Skipping fund deletion.")
            return ""

        logger.info("Deleting the account fund %s", fund_id)

        command_fund = f"mam-delete-fund -f {fund_id}"
        return self.execute_command(command_fund.split())

    def set_resource_limits(self, account: str, limits_dict: Dict[str, int]) -> str | None:
        """Set the limits for the account with the specified name."""
        if limits_dict.get("deposit", 0) < 0:
            logger.warning(
                "Skipping limit update because pricing "
                "package is not created for the related service settings."
            )
            return None

        fund_id = self._get_fund_id(account)

        if fund_id is None:
            raise exceptions.BackendError(
                "The account %s does not have a linked fund, unable to set a deposit" % account
            )

        command_deposit = f"mam-deposit -a {account} -z {limits_dict['deposit']} -f {fund_id}"
        return self.execute_command(command_deposit.split())

    def get_association(self, user: str, account: str) -> Association | None:
        """Get association between user and account."""
        command = f"mam-list-funds --raw --quiet -u {user} -a {account} --show Constraints,Balance"
        output = self.execute_command(command.split())
        lines = [line for line in output.splitlines() if "|" in line]
        if len(lines) == 0:
            return None

        return Association(account=account, user=user, value=int(lines[0].split("|")[-1]))

    def create_association(self, username: str, account: str, _: Optional[str] = None) -> str:
        """Create association between user and account in MOAB."""
        command = f"mam-modify-account --add-user +{username} -a {account}"
        return self.execute_command(command.split())

    def delete_association(self, username: str, account: str) -> str:
        """Delete association between user and account."""
        command = f"mam-modify-account --del-user {username} -a {account}"
        return self.execute_command(command.split())

    def get_usage_report(self, accounts: List[str]) -> List:
        """Get usages records from MOAB."""
        template = (
            "mam-list-usagerecords --raw --quiet --show "
            "Account,User,Charge "
            "-a %(account)s -s %(start)s -e %(end)s"
        )
        month_start, month_end = backend_utils.format_current_month()

        report_lines = []
        for account in accounts:
            command = template % {
                "account": account,
                "start": month_start,
                "end": month_end,
            }
            lines = self.execute_command(command.split()).splitlines()
            report_lines_to_add = [MoabReportLine(line) for line in lines if "|" in line]
            report_lines.extend(report_lines_to_add)

        return report_lines

    def list_account_users(self, account: str) -> List[str]:
        """Returns list of users linked to the account."""
        # TODO: make use of -A flag (fetch only active users)
        command = f"mam-list-users -a {account} --raw --show Name,DefaultAccount --quiet"
        output = self.execute_command(command.split())
        return [
            line.split("|")[0] for line in output.splitlines() if "|" in line and line[-1] != "|"
        ]



"""Functions shared between agent modules."""

import argparse
from importlib.metadata import version
from pathlib import Path
from typing import Dict, List, Set

import yaml
from waldur_client import OfferingComponent, WaldurClient, WaldurClientException

from waldur_site_agent.backends import (
    BackendType,
    logger,
)
from waldur_site_agent.backends.backend import BaseBackend, UnknownBackend
from waldur_site_agent.backends.moab_backend.backend import MoabBackend
from waldur_site_agent.backends.slurm_backend import utils as slurm_utils
from waldur_site_agent.backends.slurm_backend.backend import SlurmBackend
from waldur_site_agent.backends.structures import Resource

from . import AgentMode, Offering, WaldurAgentConfiguration

RESOURCE_ERRED_STATE = "Erred"


def init_configuration() -> WaldurAgentConfiguration:
    """Loads configuration from CLI and config file to the dataclass."""
    configuration = WaldurAgentConfiguration()
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--mode",
        "-m",
        help="Agent mode, choices: order_process, report "
        "and membership_sync; default is order_process",
        choices=["order_process", "report", "membership_sync"],
        default="order_process",
    )

    parser.add_argument(
        "--config-file",
        "-c",
        help="Path to the config file with provider settings;"
        "default is waldur-site-agent-config.yaml",
        dest="config_file_path",
        default="waldur-site-agent-config.yaml",
        required=False,
    )

    cli_args = parser.parse_args()

    config_file_path = cli_args.config_file_path
    agent_mode = cli_args.mode

    logger.info("Using %s as a config source", config_file_path)

    with Path(config_file_path).open(encoding="UTF-8") as stream:
        config = yaml.safe_load(stream)
        offering_list = config["offerings"]
        waldur_offerings = [
            Offering(
                name=offering_info["name"],
                api_url=offering_info["waldur_api_url"],
                api_token=offering_info["waldur_api_token"],
                uuid=offering_info["waldur_offering_uuid"],
                backend_type=offering_info["backend_type"].lower(),
                backend_settings=offering_info["backend_settings"],
                backend_components=offering_info["backend_components"],
            )
            for offering_info in offering_list
        ]
        configuration.waldur_offerings = waldur_offerings

        sentry_dsn = config.get("sentry_dsn")
        if sentry_dsn:
            import sentry_sdk

            sentry_sdk.init(
                dsn=sentry_dsn,
            )
            configuration.sentry_dsn = sentry_dsn

    waldur_site_agent_version = version("waldur-site-agent")

    user_agent_dict = {
        AgentMode.ORDER_PROCESS.value: "waldur-site-agent-order-process/"
        + waldur_site_agent_version,
        AgentMode.REPORT.value: "waldur-site-agent-report/" + waldur_site_agent_version,
        AgentMode.MEMBERSHIP_SYNC.value: "waldur-site-agent-membership-sync/"
        + waldur_site_agent_version,
    }

    configuration.waldur_user_agent = user_agent_dict.get(agent_mode, "")
    configuration.waldur_site_agent_mode = agent_mode
    configuration.waldur_site_agent_version = waldur_site_agent_version

    return configuration


def get_backend_for_offering(offering: Offering) -> BaseBackend:
    """Creates a corresponding backend for an offering."""
    resource_backend: BaseBackend = UnknownBackend()
    if offering.backend_type == BackendType.SLURM.value:
        resource_backend = SlurmBackend(offering.backend_settings, offering.backend_components)
    elif offering.backend_type in {
        BackendType.MOAB.value,
    }:
        resource_backend = MoabBackend(offering.backend_settings, offering.backend_components)
    elif offering.backend_type == BackendType.CUSTOM.value:
        resource_backend = UnknownBackend()
    else:
        logger.error("Unknown backend type: %s", offering.backend_type)

    return resource_backend


def delete_associations_from_waldur_allocation(
    waldur_rest_client: WaldurClient,
    backend_resource: Resource,
    usernames: Set[str],
) -> None:
    """Deletes a SLURM association for the specified resource and username in Waldur."""
    logger.info("Usernames to drop from Waldur allocation: %s", " ,".join(usernames))
    for username in usernames:
        try:
            waldur_rest_client.delete_slurm_association(backend_resource.marketplace_uuid, username)
            logger.info(
                "The user %s has been dropped from %s (backend_id: %s)",
                username,
                backend_resource.name,
                backend_resource.backend_id,
            )
        except WaldurClientException as e:
            logger.error("User %s can not be dropped due to: %s", username, e)


def create_associations_for_waldur_allocation(
    waldur_rest_client: WaldurClient,
    backend_resource: Resource,
    usernames: Set[str],
) -> None:
    """Creates a SLURM association for the specified resource and username in Waldur."""
    logger.info("New usernames to add to Waldur allocation: %s", " ,".join(usernames))
    for username in usernames:
        try:
            waldur_rest_client.create_slurm_association(backend_resource.marketplace_uuid, username)
            logger.info(
                "The user %s has been added to %s (backend_id: %s)",
                username,
                backend_resource.name,
                backend_resource.backend_id,
            )
        except WaldurClientException as e:
            logger.error("User %s can not be added due to: %s", username, e)


def mark_waldur_resources_as_erred(
    waldur_rest_client: WaldurClient,
    resources: List[Resource],
    error_details: Dict[str, str],
) -> None:
    """Marks resources in Waldur as ERRED."""
    logger.info("Marking Waldur resources as ERRED")
    for resource in resources:
        logger.info("Marking %s resource as ERRED", resource)
        try:
            waldur_rest_client.marketplace_provider_resource_set_as_erred(
                resource.marketplace_uuid, error_details
            )
        except WaldurClientException as e:
            logger.exception(
                "Waldur REST client error while setting resource state to Erred %s: %s",
                resource.backend_id,
                e,
            )


def load_offering_components() -> None:
    """Creates offering components in Waldur based on data from the config file."""
    configuration = init_configuration()
    for offering in configuration.waldur_offerings:
        logger.info("Processing %s offering", offering.name)
        waldur_rest_client = WaldurClient(
            offering.api_url, offering.api_token, configuration.waldur_user_agent
        )

        load_components_to_waldur(
            waldur_rest_client,
            offering.uuid,
            offering.name,
            offering.backend_components,
        )


def extend_backend_components(offering: Offering, waldur_offering_components: List[dict]) -> None:
    """Pulls offering component data from Waldur and populates it to the local configuration."""
    logger.info("Loading Waldur components to the local config")
    remote_components = {item["type"]: item for item in waldur_offering_components}
    missing_component_types = set(remote_components.keys()) - set(
        offering.backend_components.keys()
    )

    logger.info("Component types to add: %s", ", ".join(missing_component_types))
    for missing_component_type in missing_component_types:
        logger.info("Loading %s", missing_component_type)
        remote_component_info = remote_components[missing_component_type]
        component_info = {
            "limit": remote_component_info.get("limit_amount"),
            "measured_unit": remote_component_info["measured_unit"],
            "unit_factor": remote_component_info["unit_factor"] or 1,
            "accounting_type": remote_component_info["billing_type"],
            "label": remote_component_info["name"],
        }
        offering.backend_components[missing_component_type] = component_info


def load_components_to_waldur(
    waldur_rest_client: WaldurClient,
    offering_uuid: str,
    offering_name: str,
    components: Dict,
) -> None:
    """Creates offering components in Waldur."""
    logger.info(
        "Creating offering components data for the following resources: %s",
        ", ".join(components.keys()),
    )
    waldur_offering = waldur_rest_client.get_marketplace_provider_offering(offering_uuid)
    waldur_offering_components = {
        component["type"]: component for component in waldur_offering["components"]
    }
    for component_type, component_info in components.items():
        try:
            limit_amount = component_info.get("limit")
            accounting_type = component_info["accounting_type"]
            label = component_info["label"]

            component = OfferingComponent(
                billing_type=accounting_type,
                type=component_type,
                name=label,
                measured_unit=component_info["measured_unit"],
                limit_amount=limit_amount,
            )
            if component_type in waldur_offering_components:
                if component_info["accounting_type"] == "usage":
                    existing_component = waldur_offering_components[component_type]
                    logger.info(
                        "Offering component %s already exists, updating limit from %s to %s %s.",
                        component_type,
                        existing_component.get("limit_amount"),
                        component_info.get("limit"),
                        component_info["measured_unit"],
                    )
                    component["uuid"] = existing_component["uuid"]
                    waldur_rest_client.update_offering_component(offering_uuid, component)
                else:
                    logger.info(
                        "Offering component %s already exists, skipping creation.",
                        component_type,
                    )
            else:
                logger.info(
                    "Creating offering component %s with limit %s %s.",
                    component_type,
                    component_info["limit"],
                    component_info["measured_unit"],
                )
                waldur_rest_client.create_offering_component(offering_uuid, component)
        except Exception as e:
            logger.info(
                "Unable to create or update a component %s for offering %s (%s):",
                component_info["label"],
                offering_name,
                offering_uuid,
            )
            logger.exception(e)


def diagnostics() -> bool:
    """Performs system check for offerings."""
    configuration = init_configuration()
    logger.info("-" * 10 + "DIAGNOSTICS START" + "-" * 10)
    logger.info("Provided settings:")
    format_string = "{:<30} = {:<10}"

    if AgentMode.ORDER_PROCESS.value == configuration.waldur_site_agent_mode:
        logger.info(
            "Agent is running in %s mode - "
            "pulling orders from Waldur and creating resources in backend",
            AgentMode.ORDER_PROCESS.name,
        )
    if AgentMode.REPORT.value == configuration.waldur_site_agent_mode:
        logger.info(
            "Agent is running in %s mode - pushing usage data to Waldur",
            AgentMode.REPORT.name,
        )
    if AgentMode.MEMBERSHIP_SYNC.value == configuration.waldur_site_agent_mode:
        logger.info(
            "Agent is running in %s mode - pushing membership data to Waldur",
            AgentMode.MEMBERSHIP_SYNC.name,
        )

    for offering in configuration.waldur_offerings:
        format_string = "{:<30} = {:<10}"
        offering_uuid = offering.uuid
        offering_name = offering.name
        offering_api_url = offering.api_url
        offering_api_token = offering.api_token

        logger.info(format_string.format("Offering name", offering_name))
        logger.info(format_string.format("Offering UUID", offering_uuid))
        logger.info(format_string.format("Waldur API URL", offering_api_url))
        logger.info(format_string.format("SENTRY_DSN", str(configuration.sentry_dsn)))

        waldur_rest_client = WaldurClient(
            offering_api_url, offering_api_token, configuration.waldur_user_agent
        )

        try:
            current_user = waldur_rest_client.get_current_user()
            print_current_user(current_user)

            offering_data = waldur_rest_client.get_marketplace_provider_offering(offering_uuid)
            logger.info("Offering uuid: %s", offering_data["uuid"])
            logger.info("Offering name: %s", offering_data["name"])
            logger.info("Offering org: %s", offering_data["customer_name"])
            logger.info("Offering state: %s", offering_data["state"])

            logger.info("Offering components:")
            format_string = "{:<10} {:<10} {:<10} {:<10}"
            headers = ["Type", "Name", "Unit", "Limit"]
            logger.info(format_string.format(*headers))
            components = [
                [
                    component["type"],
                    component["name"],
                    component["measured_unit"],
                    component.get("limit_amount"),
                ]
                for component in offering_data["components"]
            ]
            for component in components:
                logger.info(format_string.format(*component))

            logger.info("")
        except WaldurClientException as err:
            logger.error("Unable to fetch offering data, reason: %s", err)

        logger.info("")
        try:
            orders = waldur_rest_client.list_orders(
                {
                    "offering_uuid": offering_uuid,
                    "state": ["pending-provider", "executing"],
                }
            )
            logger.info("Active orders:")
            format_string = "{:<10} {:<10} {:<10}"
            headers = ["Project", "Type", "State"]
            logger.info(format_string.format(*headers))
            for order in orders:
                logger.info(
                    format_string.format(order["project_name"], order["type"], order["state"])
                )
        except WaldurClientException as err:
            logger.error("Unable to fetch orders, reason: %s", err)

        backend_diagnostics_result = False
        if offering.backend_type == BackendType.SLURM.value:
            backend = SlurmBackend(offering.backend_settings, offering.backend_components)
            backend_diagnostics_result = slurm_utils.diagnostics(backend)

        if not backend_diagnostics_result:
            return False

    logger.info("-" * 10 + "DIAGNOSTICS END" + "-" * 10)
    return True


def create_homedirs_for_offering_users() -> None:
    """Creates homedirs for offering users in SLURM cluster."""
    configuration = init_configuration()
    for offering in configuration.waldur_offerings:
        # Feature is exclusive for SLURM temporarily
        if offering.backend_type != BackendType.SLURM.value or not offering.backend_settings.get(
            "enable_user_homedir_account_creation", True
        ):
            continue

        logger.info("Creating homedirs for %s offering users", offering.name)

        waldur_rest_client = WaldurClient(
            offering.api_url, offering.api_token, configuration.waldur_user_agent
        )

        offering_users = waldur_rest_client.list_remote_offering_users(
            {
                "offering_uuid": offering.uuid,
                "is_restricted": False,
            }
        )

        offering_user_usernames: Set[str] = {
            offering_user["username"] for offering_user in offering_users
        }
        umask = offering.backend_settings.get("homedir_umask", "0700")
        slurm_backend = SlurmBackend(offering.backend_settings, offering.backend_components)
        slurm_backend._create_user_homedirs(offering_user_usernames, umask)


def print_current_user(current_user: Dict) -> None:
    """Print provided user's info."""
    logger.info("Current user username: %s", current_user["username"])
    logger.info("Current user full name: %s", current_user["full_name"])
    logger.info("Current user is staff: %s", current_user.get("is_staff", False))
    logger.info("List of permissions:")
    for permission in current_user["permissions"]:
        logger.info("Role name: %s", permission["role_name"])
        logger.info("Role description: %s", permission["role_description"])
        logger.info("Scope type: %s", permission["scope_type"])
        logger.info("Scope name: %s", permission["scope_name"])
        logger.info("Scope UUID: %s", permission["scope_uuid"])
        logger.info("Expiration time: %s", permission["expiration_time"])



"""Parser module for SLURM backend."""

import datetime
from functools import cached_property
from typing import Dict

from . import utils


def parse_duration(value: str) -> float:
    """Returns duration in minutes as a float number.

    For example:
    00:01:00 is equal to 1.0
    00:00:03 is equal to 0.05
    00:01:03 is equal to 1.05
    """
    days_sep = "-"
    us_sep = "."
    simple_fmt = "%H:%M:%S"
    days_fmt = "%d-%H:%M:%S"
    us_fmt = "%H:%M:%S.%f"
    days_us_fmt = "%d-%H:%M:%S.%f"

    if days_sep not in value:
        if us_sep in value:  # Simple time with microseconds
            dt = datetime.datetime.strptime(value, us_fmt)
            delta = datetime.timedelta(
                hours=dt.hour,
                minutes=dt.minute,
                seconds=dt.second,
                microseconds=dt.microsecond,
            )
        else:  # Simple time
            dt = datetime.datetime.strptime(value, simple_fmt)
            delta = datetime.timedelta(hours=dt.hour, minutes=dt.minute, seconds=dt.second)
    elif us_sep in value:  # Simple time with microseconds and days
        dt = datetime.datetime.strptime(value, days_us_fmt)
        delta = datetime.timedelta(
            days=dt.day,
            hours=dt.hour,
            minutes=dt.minute,
            seconds=dt.second,
            microseconds=dt.microsecond,
        )
    else:  # Simple time with days
        dt = datetime.datetime.strptime(value, days_fmt)
        delta = datetime.timedelta(days=dt.day, hours=dt.hour, minutes=dt.minute, seconds=dt.second)

    return int(delta.total_seconds()) / 60


class SlurmReportLine:
    """Class for SLURM report line parsing."""

    def __init__(self, line: str, slurm_tres: Dict) -> None:
        """Inits parts field from the specified line."""
        self._parts = line.split("|")
        self.slurm_tres = slurm_tres

    @cached_property
    def account(self) -> str:
        """Returns account name from the report line."""
        return self._parts[0].strip()

    @cached_property
    def user(self) -> str:
        """Returns username from the report line."""
        return self._parts[3]

    @cached_property
    def duration(self) -> float:
        """Shortcut for parse_duration function."""
        return parse_duration(self._parts[2])

    @cached_property
    def _resources(self) -> Dict:
        pairs = self._parts[1].split(",")
        return dict(pair.split("=") for pair in pairs)

    def parse_field(self, field: str) -> int:
        """Parses integer field."""
        if field not in self._resources:
            return 0
        return utils.parse_int(self._resources[field])

    @cached_property
    def tres_usage(self) -> Dict:
        """TRES usage for the line."""
        usage = {}
        slurm_tres_set = set(self.slurm_tres.keys())
        for resource in self._resources:
            if resource not in slurm_tres_set:
                continue
            usage_raw = self.parse_field(resource)
            usage[resource] = usage_raw * self.duration
        if "mem" in usage:
            usage["mem"] = usage["mem"] // 2**20  # Convert from Bytes to MB
        return usage


class SlurmAssociationLine(SlurmReportLine):
    """Class for SLURM association line parsing."""

    @cached_property
    def user(self) -> str:
        """Always empty for association lines."""
        return ""

    @cached_property
    def duration(self) -> float:
        """Always empty for association lines."""
        return 0.0

    @cached_property
    def _resources(self) -> Dict:
        if self._parts[1] != "":
            pairs = self._parts[1].split(",")
            return dict(pair.split("=") for pair in pairs)
        return {}

    @cached_property
    def tres_limits(self) -> Dict:
        """TRES limits in the line."""
        resources = self._resources
        limits = {}
        slurm_tres_set = set(self.slurm_tres.keys())

        for tres, limit in resources.items():
            if tres not in slurm_tres_set:
                continue
            if tres == "mem":
                limits[tres] = utils.parse_int(limit) // 2**20  # Convert from Bytes to MB
            else:
                limits[tres] = utils.parse_int(limit)

        return limits



"""CLI-client for SLURM cluster."""

from __future__ import annotations

import re
from typing import Dict, List, Optional

from waldur_site_agent.backends import base
from waldur_site_agent.backends import utils as backend_utils
from waldur_site_agent.backends.structures import Account, Association

from .parser import SlurmAssociationLine, SlurmReportLine


class SlurmClient(base.BaseClient):
    """This class implements Python client for SLURM.

    See also: https://slurm.schedmd.com/sacctmgr.html
    """

    def __init__(self, slurm_tres: Dict) -> None:
        """Inits SLURM-related data."""
        self.slurm_tres = slurm_tres

    def list_accounts(self) -> List[Account]:
        """Returns a list of accounts in the SLURM cluster."""
        command = ["list", "account"]
        output = self._execute_command(command)
        return [self._parse_account(line) for line in output.splitlines() if "|" in line]

    def list_tres(self) -> List[str]:
        """Returns a list of TRES available in cluster."""
        output = self._execute_command(["list", "tres"])
        tres_list = []
        for line in output.splitlines():
            if "|" not in line:
                continue
            fields = line.split("|")
            component_type = fields[0]
            component_name = fields[1]
            if component_name:
                tres_list.append(f"{component_type}/{component_name}")
            else:
                tres_list.append(component_type)
        return tres_list

    def get_account(self, name: str) -> Account | None:
        """Returns Account object from cluster based on the account name."""
        output = self._execute_command(["show", "account", name])
        lines = [line for line in output.splitlines() if "|" in line]
        if len(lines) == 0:
            return None
        return self._parse_account(lines[0])

    def create_account(
        self,
        name: str,
        description: str,
        organization: str,
        parent_name: Optional[str] = None,
    ) -> str:
        """Creates account in the SLURM cluster."""
        parts = [
            "add",
            "account",
            name,
            f'description="{description}"',
            f'organization="{organization}"',
        ]
        if parent_name:
            parts.append(f"parent={parent_name}")
        return self._execute_command(parts)

    def delete_all_users_from_account(self, name: str) -> str:
        """Drop all the users from the account based on the account name."""
        return self._execute_command(["remove", "user", "where", f"account={name}"])

    def account_has_users(self, account: str) -> bool:
        """Checks if the account with the specified name have related users."""
        output = self._execute_command(["show", "association", "where", f"account={account}"])
        items = [self._parse_association(line) for line in output.splitlines() if "|" in line]
        return any(item.user != "" for item in items)

    def delete_account(self, name: str) -> str:
        """Deletes account with the specified name from the SLURM cluster."""
        if self.account_has_users(name):
            self.delete_all_users_from_account(name)

        return self._execute_command(["remove", "account", "where", f"name={name}"])

    def set_resource_limits(self, account: str, limits_dict: Dict[str, int]) -> str | None:
        """Sets the limits for the account with the specified name."""
        limits_str = ",".join([f"{key}={value}" for key, value in limits_dict.items()])
        quota = f"GrpTRESMins={limits_str}"
        return self._execute_command(["modify", "account", account, "set", quota])

    def set_account_qos(self, account: str, qos: str) -> None:
        """Set the specified QoS for the account."""
        self._execute_command(["modify", "account", account, "set", f"qos={qos}"])

    def unset_account_qos(self, account: str, qos: str) -> None:
        """Unset the specified QoS for the account."""
        self._execute_command(["modify", "account", account, "set", f"qos-={qos}"])

    def get_association(self, user: str, account: str) -> Association | None:
        """Returns associations between the user and the account if exists."""
        output = self._execute_command(
            [
                "show",
                "association",
                "where",
                f"user={user}",
                f"account={account}",
            ]
        )
        lines = [line for line in output.splitlines() if "|" in line]
        if len(lines) == 0:
            return None
        return self._parse_association(lines[0])

    def create_association(
        self, username: str, account: str, default_account: Optional[str] = ""
    ) -> str:
        """Creates association between the account and the user in SLURM cluster."""
        return self._execute_command(
            [
                "add",
                "user",
                username,
                f"account={account}",
                f"DefaultAccount={default_account}",
            ]
        )

    def delete_association(self, username: str, account: str) -> str:
        """Deletes association between the account and the user in SLURM cluster."""
        return self._execute_command(
            [
                "remove",
                "user",
                "where",
                f"name={username}",
                "and",
                f"account={account}",
            ]
        )

    def get_usage_report(self, accounts: List[str]) -> List:
        """Generates per-user usage report for the accounts."""
        month_start, month_end = backend_utils.format_current_month()

        args = [
            "--noconvert",
            "--truncate",
            "--allocations",
            "--allusers",
            f"--starttime={month_start}",
            f"--endtime={month_end}",
            f"--accounts={','.join(accounts)}",
            "--format=Account,ReqTRES,Elapsed,User",
        ]
        output = self._execute_command(args, "sacct", immediate=False)
        return [
            SlurmReportLine(line, self.slurm_tres) for line in output.splitlines() if "|" in line
        ]

    def get_resource_limits(self, account: str) -> List[SlurmAssociationLine]:
        """Returns limits for the account."""
        args = [
            "show",
            "association",
            "format=account,GrpTRESMins",
            "where",
            f"accounts={account}",
        ]
        output = self._execute_command(args, immediate=False)
        return [
            SlurmAssociationLine(line, self.slurm_tres)
            for line in output.splitlines()
            if "|" in line
        ]

    def list_account_users(self, account: str) -> List[str]:
        """Returns list of users linked to the account."""
        args = [
            "list",
            "associations",
            "format=account,user",
            "where",
            f"account={account}",
        ]
        output = self._execute_command(args)
        return [
            line.split("|")[1] for line in output.splitlines() if "|" in line and line[-1] != "|"
        ]

    def create_linux_user_homedir(self, username: str, umask: str = "") -> str:
        """Creates homedir for the user in Linux system."""
        return self._execute_command(
            command_name="/sbin/mkhomedir_helper",
            command=[username, umask],
            immediate=False,
            parsable=False,
        )

    def get_current_account_qos(self, account: str) -> str:
        """Returns a name of the current QoS of the account."""
        args = [
            "list",
            "associations",
            "format=account,qos",
            "where",
            f"account={account}",
        ]
        output = self._execute_command(args)
        qos_options = [
            line.split("|")[1] for line in output.splitlines() if "|" in line and line[-1] != "|"
        ]

        return qos_options[0] if len(qos_options) > 0 else ""

    def cancel_active_user_jobs(self, account: str, user: str) -> None:
        """Cancel jobs for the account and user."""
        args = [f"-u={user}", f"-A={account}", "-f"]
        self._execute_command(args, command_name="scancel")

    def list_active_user_jobs(self, account: str, user: str) -> List[str]:
        """List active jobs for the account and user."""
        args = [
            "-a",
            f"--account={account}",
            f"--user={user}",
            "--format=JobID,JobName,Partition,Account,User,State,Elapsed,Timelimit,NodeList",
        ]
        output = self._execute_command(args, command_name="sacct")
        return [line.split("|")[0] for line in output.splitlines() if "|" in line]

    def _parse_account(self, line: str) -> Account:
        parts = line.split("|")
        return Account(
            name=parts[0],
            description=parts[1],
            organization=parts[2],
        )

    def _parse_association(self, line: str) -> Association:
        parts = line.split("|")
        value = parts[9]
        match = re.match(r"cpu=(\d+)", value)
        value_ = int(match.group(1)) if match else 0
        return Association(
            account=parts[1],
            user=parts[2],
            value=value_,
        )

    def _execute_command(
        self,
        command: List[str],
        command_name: str = "sacctmgr",
        immediate: bool = True,
        parsable: bool = True,
    ) -> str:
        """Constructs and executes a command with the given parameters."""
        account_command = [command_name]
        if parsable:
            account_command.extend(["--parsable2", "--noheader"])
        if immediate:
            account_command.append("--immediate")
        account_command.extend(command)
        return self.execute_command(account_command)



"""Generic backend classes."""

from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Set, Tuple

from waldur_site_agent.backends.base import BaseClient, UnknownClient
from waldur_site_agent.backends.exceptions import BackendError

from . import BackendType, logger, structures, utils

UNKNOWN_BACKEND_TYPE = "unknown"


class BaseBackend(ABC):
    """Abstract backend class."""

    def __init__(self, backend_settings: Dict, backend_components: Dict[str, Dict]) -> None:
        """Init backend info."""
        self.backend_type = "abstract"
        self.backend_settings = backend_settings
        self.backend_components = backend_components
        self.client: BaseClient = UnknownClient()

    @abstractmethod
    def ping(self, raise_exception: bool = False) -> bool:
        """Check if backend is online."""

    @abstractmethod
    def list_components(self) -> List[str]:
        """Return a list of computing components on the backend."""

    def pull_resources(
        self, resources_info: List[structures.Resource]
    ) -> Dict[str, structures.Resource]:
        """Pull data of resources available in the backend."""
        report = {}
        for resource_info in resources_info:
            backend_id = resource_info.backend_id
            try:
                resource = self._pull_allocation(backend_id)
                if resource is not None:
                    resource.name = resource_info.name
                    resource.marketplace_uuid = resource_info.marketplace_uuid
                    resource.marketplace_scope_uuid = resource_info.marketplace_scope_uuid
                    resource.restrict_member_access = resource_info.restrict_member_access
                    resource.downscaled = resource_info.downscaled
                    resource.paused = resource_info.paused
                    resource.state = resource_info.state
                    report[backend_id] = resource
            except Exception as e:
                logger.exception("Error while pulling allocation [%s]: %s", backend_id, e)
        return report

    def _pull_allocation(self, resource_backend_id: str) -> Optional[structures.Resource]:
        """Pull allocation data from the backend."""
        account = resource_backend_id
        logger.info("Pulling allocation %s", account)
        account_info = self.client.get_account(account)

        if account_info is None:
            logger.warning("There is no %s account in the backend", account)
            return None

        users = self.client.list_account_users(account)

        report = self._get_usage_report([account])
        usage = report.get(account)

        if usage is None:
            empty_usage = {tres: 0 for tres in self.backend_components}
            usage = {"TOTAL_ACCOUNT_USAGE": empty_usage}

        return structures.Resource(
            name="",
            marketplace_uuid="",
            backend_id=account,
            limits={},
            users=users,
            usage=usage,
            backend_type=self.backend_type,
        )

    @abstractmethod
    def _get_usage_report(self, accounts: List[str]) -> Dict:
        """Collect usage report for the specified accounts."""

    def delete_resource(self, resource_backend_id: str, **kwargs: str) -> None:
        """Delete resource from the backend."""
        account = resource_backend_id

        if not account.strip():
            message = "Empty backend_id for allocation, skipping deletion"
            raise BackendError(message)

        self._delete_account_safely(account)

        if "project_slug" in kwargs:
            project_account = self._get_project_name(kwargs["project_slug"])
            if (
                len(
                    [
                        account
                        for account in self.client.list_accounts()
                        if account.organization == project_account
                        and account.name != project_account
                    ]
                )
                == 0
            ):
                self._delete_account_safely(project_account)

    def _delete_account_safely(self, account: str) -> None:
        if self.client.get_account(account):
            logger.info("Deleting the account %s from %s", account, self.backend_type)
            self.client.delete_account(account)
        else:
            logger.warning("No account %s is in %s", account, self.backend_type)

    def create_resource(self, waldur_resource: Dict) -> structures.Resource:
        """Create resource on the backend."""
        logger.info("Creating account in the backend")
        resource_uuid = waldur_resource["uuid"]
        resource_name = waldur_resource["name"]
        account_name = waldur_resource["slug"]
        project_name = waldur_resource["project_name"]
        customer_name = waldur_resource["customer_name"]

        project_account = self._get_project_name(waldur_resource["project_slug"])

        customer_account = None
        if self.backend_type == BackendType.SLURM.value:
            customer_account = self._get_customer_name(waldur_resource["customer_slug"])
            if not self.client.get_account(customer_account):
                logger.info(
                    "Creating SLURM account for customer %s (backend id = %s)",
                    customer_name,
                    customer_account,
                )
                self.client.create_account(
                    name=customer_account,
                    description=customer_name,
                    organization=customer_account,
                )

        if not self.client.get_account(project_account):
            logger.info(
                "Creating an account for project %s (backend id = %s)",
                project_name,
                project_account,
            )
            self.client.create_account(
                name=project_account,
                description=project_name,
                organization=project_account,
                parent_name=customer_account,
            )

        max_retries = 1
        account_name_generation_policy = waldur_resource["offering_plugin_options"].get(
            "account_name_generation_policy"
        )
        if account_name_generation_policy == "project_slug":
            account_name = waldur_resource["project_slug"] + "-0"
            max_retries = 10

        retries = 0
        allocation_account = self._get_allocation_name(account_name)
        while retries < max_retries:
            if self.client.get_account(allocation_account) is not None:
                logger.info(
                    "The account %s already exists in the cluster",
                    allocation_account,
                )
                if account_name_generation_policy == "project_slug":
                    # Constructing the new account name with incremented counter
                    prefix = allocation_account[:-1]
                    # The last char is guaranteed to be a valid integer
                    counter = int(allocation_account[-1]) + 1
                    allocation_account = prefix + str(counter)
            else:
                logger.info(
                    "Creating an account for allocation %s (backend id = %s)",
                    resource_name,
                    allocation_account,
                )
                self.client.create_account(
                    name=allocation_account,
                    description=resource_name,
                    organization=project_account,
                )
                break
            retries += 1

        if retries == max_retries:
            raise BackendError(
                f"Unable to create an allocation: {allocation_account} "
                "already exists in the cluster"
            )

        allocation_limits, waldur_resource_limits = self._collect_limits(waldur_resource)

        # Convert limits (for correct logging only)
        if allocation_limits:
            converted_limits = {
                key: value // self.backend_components[key].get("unit_factor", 1)
                for key, value in allocation_limits.items()
            }

            limits_str = utils.prettify_limits(converted_limits, self.backend_components)
            logger.info("Setting allocation limits to: \n%s", limits_str)
            self.client.set_resource_limits(allocation_account, allocation_limits)
        else:
            logger.info("Skipping setting of limits")

        return structures.Resource(
            backend_type=self.backend_type,
            name=resource_name,
            marketplace_uuid=resource_uuid,
            backend_id=allocation_account,
            limits=waldur_resource_limits,
        )

    @abstractmethod
    def downscale_resource(self, account: str) -> bool:
        """Downscale the account on the backend."""

    @abstractmethod
    def pause_resource(self, account: str) -> bool:
        """Pause the account on the backend."""

    @abstractmethod
    def restore_resource(self, account: str) -> bool:
        """Restore the account after downscaling or pausing."""

    @abstractmethod
    def get_resource_metadata(self, _: str) -> dict:
        """Get backend-specific resource metadata."""

    @abstractmethod
    def _collect_limits(
        self, waldur_resource: Dict[str, Dict]
    ) -> Tuple[Dict[str, int], Dict[str, int]]:
        """Collect limits for backend and waldur separately."""

    def add_users_to_resource(
        self, resource_backend_id: str, user_ids: Set[str], **kwargs: dict
    ) -> Set[str]:
        """Add specified users to the resource on the backend."""
        del kwargs
        logger.info(
            "Adding users to account %s on backend: %s", resource_backend_id, " ,".join(user_ids)
        )
        added_users = set()
        for username in user_ids:
            try:
                succeeded = self._add_user(resource_backend_id, username)
                if succeeded:
                    added_users.add(username)
            except BackendError as e:
                logger.exception(
                    "Unable to add user %s to account %s, details: %s",
                    username,
                    resource_backend_id,
                    e,
                )

        return added_users

    def _add_user(self, account: str, username: str) -> bool:
        """Add association between user and backend account if it doesn't exists."""
        if not account.strip():
            message = "Empty backend_id for allocation"
            raise BackendError(message)

        if not self.client.get_association(username, account):
            logger.info("Creating association between %s and %s", username, account)
            try:
                self.client.create_association(
                    username, account, self.backend_settings.get("default_account", "root")
                )
            except BackendError as err:
                logger.exception("Unable to create association on backend: %s", err)
                return False
        return True

    def remove_users_from_account(self, resource_backend_id: str, usernames: Set[str]) -> List[str]:
        """Remove specified users from the resource on the backend."""
        logger.info(
            "Removing users from account %s on backend: %s",
            resource_backend_id,
            " ,".join(usernames),
        )
        removed_users = []
        for username in usernames:
            try:
                succeeded = self._remove_user(resource_backend_id, username)
                if succeeded:
                    removed_users.append(username)
            except BackendError as e:
                logger.exception(
                    "Unable to remove user %s from account %s, details: %s",
                    username,
                    resource_backend_id,
                    e,
                )
        return removed_users

    def _pre_delete_user_actions(self, account: str, username: str) -> None:
        del account, username

    def _remove_user(self, account: str, username: str) -> bool:
        """Delete association between user and an account if it exists."""
        if not account.strip():
            message = "Empty account name"
            raise BackendError(message)

        if self.client.get_association(username, account):
            logger.info("Deleting association between %s and %s", username, account)
            try:
                self._pre_delete_user_actions(account, username)
                self.client.delete_association(username, account)
            except BackendError as err:
                logger.exception("Unable to delete association in the backend: %s", err)
                return False
        return True

    def set_resource_limits(self, resource_backend_id: str, limits: Dict[str, int]) -> None:
        """Set limits for the resource on the backend."""
        self.client.set_resource_limits(resource_backend_id, limits)

    def _get_allocation_name(self, allocation_slug: str, prefix: str = "") -> str:
        prefix = self.backend_settings.get("allocation_prefix", "")
        return f"{prefix}{allocation_slug}".lower()

    def _get_project_name(self, slug: str) -> str:
        return f"{self.backend_settings.get('project_prefix', '')}{slug}"

    def _get_customer_name(self, slug: str) -> str:
        return f"{self.backend_settings.get('customer_prefix', '')}{slug}"


class UnknownBackend(BaseBackend):
    """Common class for unknown backends."""

    def __init__(self) -> None:
        """Placeholder."""
        super().__init__({}, {})
        self.backend_type = UNKNOWN_BACKEND_TYPE

    def ping(self, _: bool = False) -> bool:
        """Placeholder."""
        return False

    def list_components(self) -> List[str]:
        """Placeholder."""
        return []

    def pull_resources(self, _: List[structures.Resource]) -> Dict[str, structures.Resource]:
        """Placeholder."""
        return {}

    def delete_resource(self, resource_backend_id: str, **kwargs: str) -> None:
        """Placeholder."""
        del kwargs, resource_backend_id

    def create_resource(self, _: Dict) -> structures.Resource:
        """Placeholder."""
        return structures.Resource()

    def downscale_resource(self, account: str) -> bool:
        """Placeholder."""
        del account
        return False

    def pause_resource(self, account: str) -> bool:
        """Placeholder."""
        del account
        return False

    def restore_resource(self, account: str) -> bool:
        """Placeholder."""
        del account
        return False

    def get_resource_metadata(self, _: str) -> dict:
        """Placeholder."""
        return {}

    def add_users_to_resource(
        self, resource_backend_id: str, user_ids: Set[str], **kwargs: dict
    ) -> Set[str]:
        """Placeholder."""
        del kwargs, resource_backend_id
        return user_ids

    def set_resource_limits(self, _: str, limits: Dict[str, int]) -> None:
        """Placeholder."""
        del limits

    def _collect_limits(self, _: Dict[str, Dict]) -> Tuple[Dict[str, int], Dict[str, int]]:
        return {"": 0}, {"": 0}

    def _pull_allocation(self, _: str) -> Optional[structures.Resource]:
        return None

    def _get_usage_report(self, _: List[str]) -> Dict:
        return {}



"""Init file for the main module."""

import os
from dataclasses import dataclass, field
from enum import Enum

MARKETPLACE_SLURM_OFFERING_TYPE = "Marketplace.Slurm"
WALDUR_SITE_AGENT_ORDER_PROCESS_PERIOD_MINUTES = int(
    os.environ.get("WALDUR_SITE_AGENT_ORDER_PROCESS_PERIOD_MINUTES", "5")
)
WALDUR_SITE_AGENT_REPORT_PERIOD_MINUTES = int(
    os.environ.get("WALDUR_SITE_AGENT_REPORT_PERIOD_MINUTES", "30")
)
WALDUR_SITE_AGENT_MEMBERSHIP_SYNC_PERIOD_MINUTES = int(
    os.environ.get("WALDUR_SITE_AGENT_MEMBERSHIP_SYNC_PERIOD_MINUTES", "5")
)


@dataclass
class Offering:
    """Offering structure for config file parsing."""

    name: str = ""
    api_url: str = ""
    api_token: str = ""
    uuid: str = ""
    backend_type: str = ""
    backend_settings: dict = field(default_factory=dict)
    backend_components: dict = field(default_factory=dict)


class AgentMode(Enum):
    """Enum for agent modes."""

    ORDER_PROCESS = "order_process"
    REPORT = "report"
    MEMBERSHIP_SYNC = "membership_sync"


@dataclass
class WaldurAgentConfiguration:
    """Dataclass for the agent configuration."""

    waldur_offerings: list = field(default_factory=list)
    waldur_site_agent_mode: str = ""
    waldur_user_agent: str = ""
    waldur_site_agent_version: str = ""
    sentry_dsn: str = ""



"""Generic exceptions and errors for backends."""


class BackendError(Exception):
    """Error happened on the backend."""


class ConfigurationError(Exception):
    """Agent configuration is incorrect."""



"""Module for order processing."""

from __future__ import annotations

import traceback
from time import sleep
from typing import TYPE_CHECKING, Dict, List, Set

from waldur_client import (
    SlurmAllocationState,
    is_uuid,
)

from waldur_site_agent.backends import logger
from waldur_site_agent.processors import OfferingBaseProcessor

if TYPE_CHECKING:
    from waldur_site_agent.backends.structures import Resource

from waldur_site_agent.backends.exceptions import BackendError

from . import (
    MARKETPLACE_SLURM_OFFERING_TYPE,
    WALDUR_SITE_AGENT_ORDER_PROCESS_PERIOD_MINUTES,
    Offering,
    WaldurAgentConfiguration,
    common_utils,
)


class OfferingOrderProcessor(OfferingBaseProcessor):
    """Class for an offering processing.

    Processes related orders and creates necessary associations.
    """

    def process_offering(self) -> None:
        """Pulls data form Mastermind using REST client and creates objects on the backend."""
        logger.info(
            "Processing offering %s (%s)",
            self.offering.name,
            self.offering.uuid,
        )

        self._print_current_user()

        waldur_offering = self.waldur_rest_client._get_offering(self.offering.uuid)
        common_utils.extend_backend_components(self.offering, waldur_offering["components"])

        orders = self.waldur_rest_client.list_orders(
            {
                "offering_uuid": self.offering.uuid,
                "state": ["pending-provider", "executing"],
            }
        )

        if len(orders) == 0:
            logger.info("There are no pending or executing orders")
            return

        for order in orders:
            try:
                logger.info(
                    "Processing order %s (%s) with state %s",
                    order["attributes"].get("name", "N/A"),
                    order["uuid"],
                    order["state"],
                )

                if order["state"] == "executing":
                    logger.info("Order is executing already, no need for approval")
                else:
                    logger.info("Approving the order")
                    self.waldur_rest_client.marketplace_order_approve_by_provider(order["uuid"])
                    logger.info("Refreshing the order")
                    order = self.waldur_rest_client.get_order(order["uuid"])

                if order["type"] == "Create":
                    self._process_create_order(order)

                if order["type"] == "Update":
                    self._process_update_order(order)

                if order["type"] == "Terminate":
                    self._process_terminate_order(order)

                # TODO: no need for update of orders for marketplace SLURM offerings
                logger.info("Marking order as done")
                self.waldur_rest_client.marketplace_order_set_state_done(order["uuid"])

                logger.info("The order has been successfully processed")
            except Exception as e:
                logger.exception(
                    "Error while processing order %s: %s",
                    order["uuid"],
                    e,
                )
                self.waldur_rest_client.marketplace_order_set_state_erred(
                    order["uuid"],
                    error_message=str(e),
                    error_traceback=traceback.format_exc(),
                )

    def _create_resource(
        self,
        waldur_resource: Dict,
    ) -> Resource | None:
        resource_uuid = waldur_resource["uuid"]
        resource_name = waldur_resource["name"]

        logger.info("Creating resource %s", resource_name)

        if not is_uuid(resource_uuid):
            logger.error("Unexpected resource UUID format, skipping the order")
            return None

        # TODO: figure out how to generalize it
        if (
            waldur_resource["state"] != "Creating"
            and waldur_resource["offering_type"] == MARKETPLACE_SLURM_OFFERING_TYPE
        ):
            logger.info(
                "Setting SLURM allocation state (%s) to CREATING (current state is %s)",
                waldur_resource["uuid"],
                waldur_resource["state"],
            )
            self.waldur_rest_client.set_slurm_allocation_state(
                resource_uuid, SlurmAllocationState.CREATING
            )

        backend_resource = self.resource_backend.create_resource(waldur_resource)
        if backend_resource.backend_id == "":
            msg = f"Unable to create a backend resource for offering {self.offering}"
            raise BackendError(msg)

        logger.info("Updating resource metadata in Waldur")
        self.waldur_rest_client.marketplace_provider_resource_set_backend_id(
            resource_uuid, backend_resource.backend_id
        )

        if waldur_resource["offering_type"] == MARKETPLACE_SLURM_OFFERING_TYPE:
            logger.info("Setting SLURM allocation backend ID")
            self.waldur_rest_client.set_slurm_allocation_backend_id(
                waldur_resource["uuid"], backend_resource.backend_id
            )

            logger.info("Updating allocation limits in Waldur")
            self.waldur_rest_client.set_slurm_allocation_limits(
                waldur_resource["uuid"], backend_resource.limits
            )

        return backend_resource

    def _add_users_to_resource(
        self,
        backend_resource: Resource,
    ) -> None:
        logger.info("Adding users to resource")
        logger.info("Fetching Waldur resource team")
        team = self.waldur_rest_client.marketplace_provider_resource_get_team(
            backend_resource.marketplace_uuid
        )
        user_uuids = {user["uuid"] for user in team}

        logger.info("Fetching Waldur offering users")
        offering_users_all = self.waldur_rest_client.list_remote_offering_users(
            {"offering_uuid": self.offering.uuid, "is_restricted": False}
        )
        offering_usernames: Set[str] = {
            offering_user["username"]
            for offering_user in offering_users_all
            if offering_user["user_uuid"] in user_uuids and offering_user["username"] != ""
        }

        logger.info("Adding usernames to resource in backend")
        added_users = self.resource_backend.add_users_to_resource(
            backend_resource.backend_id,
            offering_usernames,
            homedir_umask=self.offering.backend_settings.get("homedir_umask", "0700"),
        )

        common_utils.create_associations_for_waldur_allocation(
            self.waldur_rest_client, backend_resource, added_users
        )

    def _process_create_order(self, order: Dict) -> None:
        # Wait until resource is created
        attempts = 0
        max_attempts = 4
        while "marketplace_resource_uuid" not in order:
            if attempts > max_attempts:
                logger.error("Order processing timed out")
                return

            if order["state"] != "executing":
                logger.error("order has unexpected state %s", order["state"])
                return

            logger.info("Waiting for resource creation...")
            sleep(5)

            order = self.waldur_rest_client.get_order(order["uuid"])
            attempts += 1

        if order["offering_type"] == MARKETPLACE_SLURM_OFFERING_TYPE:
            # TODO: drop this cycle
            # after removal of waldur_slurm.Allocation model from Mastermind
            attempts = 0
            while order["resource_uuid"] is None:
                if attempts > max_attempts:
                    logger.error("Order processing timed out")
                    return

                if order["state"] != "executing":
                    logger.error("order has unexpected state %s", order["state"])
                    return

                logger.info("Waiting for Waldur allocation creation...")
                sleep(5)

                order = self.waldur_rest_client.get_order(order["uuid"])
                attempts += 1

        waldur_resource = self.waldur_rest_client.get_marketplace_provider_resource(
            order["marketplace_resource_uuid"]
        )

        waldur_resource["project_slug"] = order["project_slug"]
        waldur_resource["customer_slug"] = order["customer_slug"]

        backend_resource = self._create_resource(waldur_resource)
        if backend_resource is None:
            msg = "Unable to create a resource"
            raise BackendError(msg)

        if order["offering_type"] == MARKETPLACE_SLURM_OFFERING_TYPE:
            logger.info("Updating Waldur resource scope state")
            self.waldur_rest_client.set_slurm_allocation_state(
                waldur_resource["uuid"], SlurmAllocationState.OK
            )

            self._add_users_to_resource(
                backend_resource,
            )

    def _process_update_order(self, order: dict) -> None:
        logger.info("Updating limits for %s", order["resource_name"])
        resource_uuid = order["marketplace_resource_uuid"]
        waldur_resource = self.waldur_rest_client.get_marketplace_provider_resource(resource_uuid)

        if order["offering_type"] == MARKETPLACE_SLURM_OFFERING_TYPE:
            self.waldur_rest_client.set_slurm_allocation_state(
                resource_uuid, SlurmAllocationState.UPDATING
            )

        resource_backend = common_utils.get_backend_for_offering(self.offering)
        if resource_backend is None:
            return

        waldur_resource_backend_id = waldur_resource["backend_id"]

        new_limits = order["limits"]
        if not new_limits:
            logger.error(
                "Order %s (resource %s) with type" + "Update does not include new limits",
                order["uuid"],
                waldur_resource["name"],
            )

        if new_limits:
            resource_backend.set_resource_limits(waldur_resource_backend_id, new_limits)

        if order["offering_type"] == MARKETPLACE_SLURM_OFFERING_TYPE:
            logger.info("Updating Waldur resource scope state")
            self.waldur_rest_client.set_slurm_allocation_state(
                resource_uuid, SlurmAllocationState.OK
            )

        logger.info(
            "The limits for %s were updated successfully from %s to %s",
            waldur_resource["name"],
            order["attributes"]["old_limits"],
            new_limits,
        )

    def _process_terminate_order(self, order: dict) -> None:
        logger.info("Terminating resource %s", order["resource_name"])
        resource_uuid = order["marketplace_resource_uuid"]

        waldur_resource = self.waldur_rest_client.get_marketplace_provider_resource(resource_uuid)
        project_slug = order["project_slug"]

        resource_backend = common_utils.get_backend_for_offering(self.offering)
        if resource_backend is None:
            return

        resource_backend.delete_resource(waldur_resource["backend_id"], project_slug=project_slug)

        logger.info("Allocation has been terminated successfully")


def process_offerings(waldur_offerings: List[Offering], user_agent: str = "") -> None:
    """Processes offerings one-by-one."""
    logger.info("Number of offerings to process: %s", len(waldur_offerings))
    for offering in waldur_offerings:
        try:
            processor = OfferingOrderProcessor(offering, user_agent)
            processor.process_offering()
        except Exception as e:
            logger.exception("The application crashed due to the error: %s", e)


def start(configuration: WaldurAgentConfiguration) -> None:
    """Starts the main loop for offering processing."""
    logger.info("Synching data from Waldur")
    while True:
        try:
            process_offerings(configuration.waldur_offerings, configuration.waldur_user_agent)
        except Exception as e:
            logger.exception("The application crashed due to the error: %s", e)
        sleep(WALDUR_SITE_AGENT_ORDER_PROCESS_PERIOD_MINUTES * 60)



"""Main application module."""

from waldur_site_agent.backends import logger

from . import AgentMode, agent_membership_sync, agent_order_process, agent_report, common_utils


def main() -> None:
    """Entrypoint for the application."""
    configuration = common_utils.init_configuration()
    logger.info(
        "Waldur site Agent version: %s, site: SLURM", configuration.waldur_site_agent_version
    )

    logger.info("Running agent in %s mode", configuration.waldur_site_agent_mode)
    if AgentMode.ORDER_PROCESS.value == configuration.waldur_site_agent_mode:
        agent_order_process.start(configuration)
    if AgentMode.REPORT.value == configuration.waldur_site_agent_mode:
        agent_report.start(configuration)
    if AgentMode.MEMBERSHIP_SYNC.value == configuration.waldur_site_agent_mode:
        agent_membership_sync.start(configuration)


if __name__ == "__main__":
    main()



"""Parsing classes for MOAB."""

import decimal
from functools import cached_property
from typing import Dict


class MoabReportLine:
    """Parser for report lines from MOAB."""

    def __init__(self, line: str) -> None:
        """Constructor."""
        self._parts = line.split("|")

    @cached_property
    def account(self) -> str:
        """Parses account name."""
        return self._parts[0].strip()

    @cached_property
    def user(self) -> str:
        """Parses username."""
        return self._parts[1]

    @cached_property
    def charge(self) -> decimal.Decimal:
        """Parses charge."""
        return decimal.Decimal(self._parts[2])

    @cached_property
    def usages(self) -> Dict[str, int]:
        """Return deposit usage from the report line."""
        return {"deposit": int(self.charge)}



"""SLURM-specific backend classes and functions."""

from __future__ import annotations

from typing import Dict, List, Set, Tuple

from waldur_site_agent.backends import (
    BackendType,
    backend,
    logger,
)
from waldur_site_agent.backends import utils as backend_utils
from waldur_site_agent.backends.exceptions import BackendError

from . import utils
from .client import SlurmClient


class SlurmBackend(backend.BaseBackend):
    """Main class for management of SLURM resources."""

    def __init__(self, slurm_settings: Dict, slurm_tres: Dict[str, Dict]) -> None:
        """Init backend data and creates a corresponding client."""
        super().__init__(slurm_settings, slurm_tres)
        self.backend_type = BackendType.SLURM.value
        self.client: SlurmClient = SlurmClient(slurm_tres)

    def ping(self, raise_exception: bool = False) -> bool:
        """Check if the SLURM cluster is online."""
        try:
            self.client.list_accounts()
        except BackendError as err:
            if raise_exception:
                raise
            logger.info("Error: %s", err)
            return False
        else:
            return True

    def list_components(self) -> List[str]:
        """Return a list of TRES on the SLURM cluster."""
        return self.client.list_tres()

    def _collect_limits(
        self, waldur_resource: Dict[str, Dict]
    ) -> Tuple[Dict[str, int], Dict[str, int]]:
        """Collect SLURM and Waldur limits separately."""
        allocation_limits = backend_utils.get_usage_based_limits(self.backend_components)
        limit_based_components = [
            component
            for component, data in self.backend_components.items()
            if data["accounting_type"] == "limit"
        ]

        # Add limit-based limits
        for component_key in limit_based_components:
            allocation_limits[component_key] = (
                waldur_resource["limits"][component_key]
                * self.backend_components[component_key]["unit_factor"]
            )

        # Keep only limit-based components for Waldur resource
        waldur_resource_limits = {
            component_key: waldur_resource["limits"][component_key]
            for component_key, data in self.backend_components.items()
            if data["accounting_type"] == "limit"
        }

        return allocation_limits, waldur_resource_limits

    def add_users_to_resource(
        self, resource_backend_id: str, user_ids: Set[str], **kwargs: dict
    ) -> Set[str]:
        """Add specified users to the allocations on the SLURM cluster."""
        added_users = super().add_users_to_resource(resource_backend_id, user_ids)

        if self.backend_settings.get("enable_user_homedir_account_creation", True):
            umask: str = str(kwargs.get("homedir_umask", "0700"))
            self._create_user_homedirs(added_users, umask=umask)

        return added_users

    def downscale_resource(self, account: str) -> bool:
        """Downscale the resource QoS respecting the backend settings."""
        qos_downscaled = self.backend_settings.get("qos_downscaled")
        if not qos_downscaled:
            logger.error(
                "The QoS for dowscaling has incorrect value %s, skipping operation",
                qos_downscaled,
            )
            return False

        current_qos = self.client.get_current_account_qos(account)

        logger.info("Current QoS: %s", current_qos)

        if current_qos == qos_downscaled:
            logger.info("The account is already downscaled")
            return True

        logger.info("Setting %s QoS for the SLURM account", qos_downscaled)
        self.client.set_account_qos(account, qos_downscaled)
        logger.info("The new QoS successfully set")
        return True

    def pause_resource(self, account: str) -> bool:
        """Set the resource QoS to a paused one respecting the backend settings."""
        qos_paused = self.backend_settings.get("qos_paused")
        if not qos_paused:
            logger.error(
                "The QoS for pausing has incorrect value %s, skipping operation",
                qos_paused,
            )
            return False

        current_qos = self.client.get_current_account_qos(account)

        logger.info("Current QoS: %s", current_qos)

        if current_qos == qos_paused:
            logger.info("The account is already paused")
            return True

        logger.info("Setting %s QoS for the SLURM account", qos_paused)
        self.client.set_account_qos(account, qos_paused)
        logger.info("The new QoS successfully set")
        return True

    def restore_resource(self, account: str) -> bool:
        """Unset the resource QoS."""
        current_qos = self.client.get_current_account_qos(account)

        if current_qos in [None, ""]:
            logger.info("The account does not have an active QoS set, skipping reset")
            return False

        logger.info("The current QoS is %s", current_qos)

        if current_qos == "normal":
            logger.info("The account already has normal QoS")
            return False

        qos_paused = self.backend_settings.get("qos_paused")
        qos_downscaled = self.backend_settings.get("qos_downscaled")

        if current_qos not in (qos_paused, qos_downscaled):
            logger.info(
                "The current QoS does not match either downscaled (%s) or paused (%s)",
                qos_downscaled,
                qos_paused,
            )
            return False

        logger.info("Unset %s QoS", current_qos)
        self.client.unset_account_qos(account, current_qos)
        new_qos = self.client.get_current_account_qos(account)
        logger.info("The new QoS is %s", new_qos)

        return True

    def get_resource_metadata(self, account: str) -> dict:
        """Return backend metadata for the SLURM account (QoS only for now)."""
        current_qos = self.client.get_current_account_qos(account)
        return {"qos": current_qos}

    def _create_user_homedirs(self, usernames: Set[str], umask: str = "0700") -> None:
        logger.info("Creating homedirs for users")
        for username in usernames:
            try:
                self.client.create_linux_user_homedir(username, umask)
                logger.info("Homedir for user %s has been created", username)
            except BackendError as err:
                logger.exception(
                    "Unable to create user homedir for %s, reason: %s",
                    username,
                    err,
                )

    def _get_usage_report(self, accounts: List[str]) -> Dict[str, Dict[str, Dict[str, int]]]:
        """Example output.

        {
            "account_name": {
                "TOTAL_ACCOUNT_USAGE": {
                    'cpu': 1,
                    'gres/gpu': 2,
                    'mem': 3,
                },
                "user1": {
                    'cpu': 1,
                    'gres/gpu': 2,
                    'mem': 3,
                },
            }
        }
        """
        report: Dict[str, Dict[str, Dict[str, int]]] = {}
        lines = self.client.get_usage_report(accounts)

        for line in lines:
            report.setdefault(line.account, {}).setdefault(line.user, {})
            tres_usage = line.tres_usage
            user_usage_existing = report[line.account][line.user]
            user_usage_new = backend_utils.sum_dicts([user_usage_existing, tres_usage])
            report[line.account][line.user] = user_usage_new

        for account_usage in report.values():
            usages_per_user = list(account_usage.values())
            total = backend_utils.sum_dicts(usages_per_user)
            account_usage["TOTAL_ACCOUNT_USAGE"] = total

        # Convert SLURM units to Waldur ones
        report_converted: Dict[str, Dict[str, Dict[str, int]]] = {}
        for account, account_usage in report.items():
            report_converted[account] = {}
            for username, usage_dict in account_usage.items():
                converted_usage_dict = utils.convert_slurm_units_to_waldur_ones(
                    self.backend_components, usage_dict
                )
                report_converted[account][username] = converted_usage_dict

        return report_converted

    def list_active_user_jobs(self, account: str, user: str) -> List[str]:
        """List active jobs for account and user."""
        logger.info("Listing jobs for account %s and user %s", account, user)
        return self.client.list_active_user_jobs(account, user)

    def cancel_active_jobs_for_account_user(self, account: str, user: str) -> None:
        """Cancel account the active jobs for the specified account and user."""
        logger.info("Cancelling jobs for the account %s and user %s", account, user)
        self.client.cancel_active_user_jobs(account, user)

    def _pre_delete_user_actions(self, account: str, username: str) -> None:
        job_ids = self.list_active_user_jobs(account, username)
        if len(job_ids) > 0:
            logger.info(
                "The active jobs for account %s and user %s: %s",
                account,
                username,
                ", ".join(job_ids),
            )
            self.cancel_active_jobs_for_account_user(account, username)

    def _get_allocation_limits(self, account: str) -> Dict[str, int]:
        """Return limits converted to Waldur-readable values."""
        lines = self.client.get_resource_limits(account)
        correct_lines = [
            association.tres_limits for association in lines if association.tres_limits
        ]
        if len(correct_lines) == 0:
            return {}

        return utils.convert_slurm_units_to_waldur_ones(
            self.backend_components, correct_lines[0], to_int=True
        )

    def set_resource_limits(self, resource_backend_id: str, limits: Dict[str, int]) -> None:
        """Set limits for limit-based components in the SLURM allocation."""
        # Convert limits
        converted_limits = {
            key: value * self.backend_components[key]["unit_factor"]
            for key, value in limits.items()
        }
        super().set_resource_limits(resource_backend_id, converted_limits)
